@@ -267,6 +267,8 @@ struct DbObject {
     struct EnvObject *env; /* Not refcounted. */
     /** MDB database handle. */
     MDB_dbi dbi;
+    /** Flags at time of creation. */
+    unsigned int flags;
 };
 
 /** lmdb.Environment */
@@ -325,6 +327,8 @@ struct CursorObject {
     /** If TransObject.mutations!=last_mutation, must MDB_GET_CURRENT to
      * refresh `key' and `val'. */
     int last_mutation;
+    /** DBI flags at time of creation. */
+    unsigned int dbi_flags;
 };
 
 
@@ -969,6 +973,7 @@ make_cursor(DbObject *db, TransObject *trans)
     self->val.mv_size = 0;
     self->trans = trans;
     self->last_mutation = trans->mutations;
+    self->dbi_flags = db->flags;
     Py_INCREF(self->trans);
     return (PyObject *) self;
 }
@@ -983,6 +988,7 @@ db_from_name(EnvObject *env, MDB_txn *txn, const char *name,
              unsigned int flags)
 {
     MDB_dbi dbi;
+    unsigned int f;
     int rc;
     DbObject *dbo;
 
@@ -991,6 +997,11 @@ db_from_name(EnvObject *env, MDB_txn *txn, const char *name,
         err_set("mdb_dbi_open", rc);
         return NULL;
     }
+    if((rc = mdb_dbi_flags(txn, dbi, &f))) {
+        err_set("mdb_dbi_flags", rc);
+        mdb_dbi_close(env, dbi);
+        return NULL;
+    }
 
     if(! ((dbo = PyObject_New(DbObject, &PyDatabase_Type)))) {
         return NULL;
@@ -1000,6 +1011,7 @@ db_from_name(EnvObject *env, MDB_txn *txn, const char *name,
     LINK_CHILD(env, dbo)
     dbo->env = env; /* no refcount */
     dbo->dbi = dbi;
+    dbo->flags = f;
     DEBUG("DbObject '%s' opened at %p", name, dbo)
     return dbo;
 }
@@ -1055,7 +1067,6 @@ db_clear(DbObject *self)
 static PyObject *
 db_flags(DbObject *self, PyObject *args, PyObject *kwds)
 {
-    unsigned int f;
     PyObject *dct;
     int rc;
 
@@ -1076,12 +1087,9 @@ db_flags(DbObject *self, PyObject *args, PyObject *kwds)
     if(! arg.txn->valid) {
         return err_invalid();
     }
-    if((rc = mdb_dbi_flags(arg.txn->txn, self->dbi, &f))) {
-        err_set("mdb_dbi_flags", rc);
-        return NULL;
-    }
 
     dct = PyDict_New();
+    unsigned int f = self->flags;
     PyDict_SetItemString(dct, "reverse_key", py_bool(f & MDB_REVERSEKEY));
     PyDict_SetItemString(dct, "dupsort", py_bool(f & MDB_DUPSORT));
     return dct;
@@ -1822,6 +1830,9 @@ _cursor_get_c(CursorObject *self, enum MDB_cursor_op op)
 static PyObject *
 _cursor_get(CursorObject *self, enum MDB_cursor_op op)
 {
+    if(! self->valid) {
+        return err_invalid();
+    }
     if(_cursor_get_c(self, op)) {
         return NULL;
     }
@@ -1829,30 +1840,39 @@ _cursor_get(CursorObject *self, enum MDB_cursor_op op)
 }
 
 /**
- * Cursor.delete() -> bool
+ * Cursor.delete(dupdata=False) -> bool
  */
 static PyObject *
-cursor_delete(CursorObject *self)
+cursor_delete(CursorObject *self, PyObject *args, PyObject *kwds)
 {
-    if(! self->valid) {
-        return err_invalid();
+    struct cursor_delete {
+        int dupdata;
+    } arg = {0};
+
+    static const struct argspec argspec[] = {
+        {ARG_BOOL, DUPDATA_S, OFFSET(cursor_delete, dupdata)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
     }
-    PyObject *res = Py_False;
+
+    int res = 0;
     if(self->positioned) {
         DEBUG("deleting key '%.*s'",
               (int) self->key.mv_size,
               (char*) self->key.mv_data)
         int rc;
-        UNLOCKED(rc, mdb_cursor_del(self->curs, 0));
+        int flags = arg.dupdata ? MDB_NODUPDATA : 0;
+        UNLOCKED(rc, mdb_cursor_del(self->curs, flags));
         self->trans->mutations++;
         if(rc) {
             return err_set("mdb_cursor_del", rc);
         }
-        res = Py_True;
+        res = 1;
         _cursor_get_c(self, MDB_GET_CURRENT);
     }
-    Py_INCREF(res);
-    return res;
+    return py_bool(res);
 }
 
 /**
@@ -1861,12 +1881,18 @@ cursor_delete(CursorObject *self)
 static PyObject *
 cursor_first(CursorObject *self)
 {
-    if(! self->valid) {
-        return err_invalid();
-    }
     return _cursor_get(self, MDB_FIRST);
 }
 
+/**
+ * Cursor.first_dup() -> bool
+ */
+static PyObject *
+cursor_first_dup(CursorObject *self)
+{
+    return _cursor_get(self, MDB_FIRST_DUP);
+}
+
 static PyObject *
 cursor_value(CursorObject *self);
 
@@ -1962,36 +1988,72 @@ cursor_key(CursorObject *self)
 static PyObject *
 cursor_last(CursorObject *self)
 {
-    if(! self->valid) {
-        return err_invalid();
-    }
     return _cursor_get(self, MDB_LAST);
 }
 
+/**
+ * Cursor.last_dup() -> bool
+ */
+static PyObject *
+cursor_last_dup(CursorObject *self)
+{
+    return _cursor_get(self, MDB_LAST_DUP);
+}
+
 /**
  * Cursor.next() -> bool
  */
 static PyObject *
 cursor_next(CursorObject *self)
 {
-    if(! self->valid) {
-        return err_invalid();
-    }
     return _cursor_get(self, MDB_NEXT);
 }
 
+/**
+ * Cursor.next_dup() -> bool
+ */
+static PyObject *
+cursor_next_dup(CursorObject *self)
+{
+    return _cursor_get(self, MDB_NEXT_DUP);
+}
+
+/**
+ * Cursor.next_nodup() -> bool
+ */
+static PyObject *
+cursor_next_nodup(CursorObject *self)
+{
+    return _cursor_get(self, MDB_NEXT_NODUP);
+}
+
 /**
  * Cursor.prev() -> bool
  */
 static PyObject *
 cursor_prev(CursorObject *self)
 {
-    if(! self->valid) {
-        return err_invalid();
-    }
     return _cursor_get(self, MDB_PREV);
 }
 
+/**
+ * Cursor.prev_dup() -> bool
+ */
+static PyObject *
+cursor_prev_dup(CursorObject *self)
+{
+    return _cursor_get(self, MDB_PREV_DUP);
+}
+
+/**
+ * Cursor.prev_nodup() -> bool
+ */
+static PyObject *
+cursor_prev_nodup(CursorObject *self)
+{
+    return _cursor_get(self, MDB_PREV_NODUP);
+}
+
 /**
  * Cursor.put() -> bool
  */
@@ -2004,7 +2066,7 @@ cursor_put(CursorObject *self, PyObject *args, PyObject *kwds)
         int dupdata;
         int overwrite;
         int append;
-    } arg = {{0, 0}, {0, 0}, 0, 1, 0};
+    } arg = {{0, 0}, {0, 0}, 1, 1, 0};
 
     static const struct argspec argspec[] = {
         {ARG_BUF, KEY_S, OFFSET(cursor_put, key)},
@@ -2041,6 +2103,59 @@ cursor_put(CursorObject *self, PyObject *args, PyObject *kwds)
     Py_RETURN_TRUE;
 }
 
+/**
+ * Shared between Cursor.replace() and Transaction.replace()
+ */
+static PyObject *
+do_cursor_replace(CursorObject *self, MDB_val *key, MDB_val *val)
+{
+    int rc;
+    PyObject *old;
+    MDB_val newval = *val;
+
+    if(self->dbi_flags & MDB_DUPSORT) {
+        self->key = *key;
+        if(_cursor_get_c(self, MDB_SET_KEY)) {
+            return NULL;
+        }
+        if(self->positioned) {
+            if(! ((old = obj_from_val(&self->val, 0)))) {
+                return NULL;
+            }
+            UNLOCKED(rc, mdb_cursor_del(self->curs, MDB_NODUPDATA));
+            self->trans->mutations++;
+            if(rc) {
+                Py_CLEAR(old);
+                return err_set("mdb_cursor_del", rc);
+            }
+        } else {
+            old = Py_None;
+            Py_INCREF(old);
+        }
+    } else {
+        /* val is updated if MDB_KEYEXIST. */
+        int flags = MDB_NOOVERWRITE;
+        UNLOCKED(rc, mdb_cursor_put(self->curs, key, val, flags));
+        self->trans->mutations++;
+        if(! rc) {
+            Py_RETURN_NONE;
+        } else if(rc != MDB_KEYEXIST) {
+            return err_set("mdb_put", rc);
+        }
+
+        if(! ((old = obj_from_val(val, 0)))) {
+            return NULL;
+        }
+    }
+
+    UNLOCKED(rc, mdb_cursor_put(self->curs, key, &newval, 0));
+    if(rc) {
+        Py_DECREF(old);
+        return err_set("mdb_put", rc);
+    }
+    return old;
+}
+
 /**
  * Cursor.replace() -> None|result
  */
@@ -2061,30 +2176,7 @@ cursor_replace(CursorObject *self, PyObject *args, PyObject *kwds)
         return NULL;
     }
 
-    /* arg.val is updated if MDB_KEYEXIST. */
-    MDB_val newval = arg.val;
-    int flags = MDB_NOOVERWRITE;
-    int rc;
-    UNLOCKED(rc, mdb_cursor_put(self->curs, &arg.key, &arg.val, flags));
-    self->trans->mutations++;
-    if(! rc) {
-        Py_RETURN_NONE;
-    } else if(rc != MDB_KEYEXIST) {
-        return err_set("mdb_put", rc);
-    }
-
-    PyObject *old = obj_from_val(&arg.val, 0);
-    if(! old) {
-        return NULL;
-    }
-
-    UNLOCKED(rc, mdb_cursor_put(self->curs, &arg.key, &newval, 0));
-    if(rc) {
-        Py_DECREF(old);
-        return err_set("mdb_put", rc);
-    }
-
-    return old;
+    return do_cursor_replace(self, &arg.key, &arg.val);
 }
 
 /**
@@ -2128,7 +2220,7 @@ cursor_pop(CursorObject *self, PyObject *args, PyObject *kwds)
 }
 
 /**
- * Cursor.set_key() -> bool
+ * Cursor.set_key(key) -> bool
  */
 static PyObject *
 cursor_set_key(CursorObject *self, PyObject *arg)
@@ -2143,7 +2235,31 @@ cursor_set_key(CursorObject *self, PyObject *arg)
 }
 
 /**
- * Cursor.set_range() -> bool
+ * Cursor.set_key_dup(key, value) -> bool
+ */
+static PyObject *
+cursor_set_key_dup(CursorObject *self, PyObject *args, PyObject *kwds)
+{
+    struct cursor_set_key_dup {
+        MDB_val key;
+        MDB_val value;
+    } arg = {{0, 0}, {0, 0}};
+
+    static const struct argspec argspec[] = {
+        {ARG_BUF, KEY_S, OFFSET(cursor_set_key_dup, key)},
+        {ARG_BUF, VALUE_S, OFFSET(cursor_set_key_dup, value)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+    self->key = arg.key;
+    self->val = arg.value;
+    return _cursor_get(self, MDB_GET_BOTH);
+}
+
+/**
+ * Cursor.set_range(key) -> bool
  */
 static PyObject *
 cursor_set_range(CursorObject *self, PyObject *arg)
@@ -2160,6 +2276,30 @@ cursor_set_range(CursorObject *self, PyObject *arg)
     return _cursor_get(self, MDB_FIRST);
 }
 
+/**
+ * Cursor.set_range_dup(key, value) -> bool
+ */
+static PyObject *
+cursor_set_range_dup(CursorObject *self, PyObject *args, PyObject *kwds)
+{
+    struct cursor_set_range_dup {
+        MDB_val key;
+        MDB_val value;
+    } arg = {{0, 0}, {0, 0}};
+
+    static const struct argspec argspec[] = {
+        {ARG_BUF, KEY_S, OFFSET(cursor_set_range_dup, key)},
+        {ARG_BUF, VALUE_S, OFFSET(cursor_set_range_dup, value)}
+    };
+
+    if(parse_args(self->valid, SPECSIZE(), argspec, args, kwds, &arg)) {
+        return NULL;
+    }
+    self->key = arg.key;
+    self->val = arg.value;
+    return _cursor_get(self, MDB_GET_BOTH);
+}
+
 /**
  * Cursor.value() -> result
  */
@@ -2193,12 +2333,13 @@ new_iterator(CursorObject *cursor, IterValFunc val_func, MDB_cursor_op op)
 
 static PyObject *
 iter_from_args(CursorObject *self, PyObject *args, PyObject *kwds,
-               enum MDB_cursor_op pos_op, enum MDB_cursor_op op)
+               signed int pos_op, enum MDB_cursor_op op,
+               int keys_default, int values_default)
 {
     struct iter_from_args {
         int keys;
         int values;
-    } arg = {1, 1};
+    } arg = {keys_default, values_default};
 
     static const struct argspec argspec[] = {
         {ARG_BOOL, KEYS_S, OFFSET(iter_from_args, keys)},
@@ -2209,8 +2350,8 @@ iter_from_args(CursorObject *self, PyObject *args, PyObject *kwds,
         return NULL;
     }
 
-    if(! self->positioned) {
-        if(_cursor_get_c(self, pos_op)) {
+    if(pos_op != -1 && !self->positioned) {
+        if(_cursor_get_c(self, (enum MDB_cursor_op) pos_op)) {
             return NULL;
         }
     }
@@ -2229,7 +2370,7 @@ iter_from_args(CursorObject *self, PyObject *args, PyObject *kwds,
 static PyObject *
 cursor_iter(CursorObject *self)
 {
-    return iter_from_args(self, NULL, NULL, MDB_FIRST, MDB_NEXT);
+    return iter_from_args(self, NULL, NULL, MDB_FIRST, MDB_NEXT, 1, 1);
 }
 
 /**
@@ -2238,7 +2379,25 @@ cursor_iter(CursorObject *self)
 static PyObject *
 cursor_iternext(CursorObject *self, PyObject *args, PyObject *kwargs)
 {
-    return iter_from_args(self, args, kwargs, MDB_FIRST, MDB_NEXT);
+    return iter_from_args(self, args, kwargs, MDB_FIRST, MDB_NEXT, 1, 1);
+}
+
+/**
+ * Cursor.iternext_dup() -> Iterator
+ */
+static PyObject *
+cursor_iternext_dup(CursorObject *self, PyObject *args, PyObject *kwargs)
+{
+    return iter_from_args(self, args, kwargs, -1, MDB_NEXT_DUP, 0, 1);
+}
+
+/**
+ * Cursor.iternext_nodup() -> Iterator
+ */
+static PyObject *
+cursor_iternext_nodup(CursorObject *self, PyObject *args, PyObject *kwargs)
+{
+    return iter_from_args(self, args, kwargs, MDB_FIRST, MDB_NEXT_NODUP, 1, 0);
 }
 
 /**
@@ -2247,7 +2406,25 @@ cursor_iternext(CursorObject *self, PyObject *args, PyObject *kwargs)
 static PyObject *
 cursor_iterprev(CursorObject *self, PyObject *args, PyObject *kwargs)
 {
-    return iter_from_args(self, args, kwargs, MDB_LAST, MDB_PREV);
+    return iter_from_args(self, args, kwargs, MDB_LAST, MDB_PREV, 1, 1);
+}
+
+/**
+ * Cursor.iterprev_dup() -> Iterator
+ */
+static PyObject *
+cursor_iterprev_dup(CursorObject *self, PyObject *args, PyObject *kwargs)
+{
+    return iter_from_args(self, args, kwargs, -1, MDB_PREV_DUP, 0, 1);
+}
+
+/**
+ * Cursor.iterprev_nodup() -> Iterator
+ */
+static PyObject *
+cursor_iterprev_nodup(CursorObject *self, PyObject *args, PyObject *kwargs)
+{
+    return iter_from_args(self, args, kwargs, MDB_LAST, MDB_PREV_NODUP, 1, 0);
 }
 
 /**
@@ -2298,21 +2475,33 @@ cursor_iter_from(CursorObject *self, PyObject *args)
 
 static struct PyMethodDef cursor_methods[] = {
     {"count", (PyCFunction)cursor_count, METH_NOARGS},
-    {"delete", (PyCFunction)cursor_delete, METH_NOARGS},
+    {"delete", (PyCFunction)cursor_delete, METH_VARARGS|METH_KEYWORDS},
     {"first", (PyCFunction)cursor_first, METH_NOARGS},
+    {"first_dup", (PyCFunction)cursor_first_dup, METH_NOARGS},
     {"get", (PyCFunction)cursor_get, METH_VARARGS|METH_KEYWORDS},
     {"item", (PyCFunction)cursor_item, METH_NOARGS},
     {"iternext", (PyCFunction)cursor_iternext, METH_VARARGS|METH_KEYWORDS},
+    {"iternext_dup", (PyCFunction)cursor_iternext_dup, METH_VARARGS|METH_KEYWORDS},
+    {"iternext_nodup", (PyCFunction)cursor_iternext_nodup, METH_VARARGS|METH_KEYWORDS},
     {"iterprev", (PyCFunction)cursor_iterprev, METH_VARARGS|METH_KEYWORDS},
+    {"iterprev_dup", (PyCFunction)cursor_iterprev_dup, METH_VARARGS|METH_KEYWORDS},
+    {"iterprev_nodup", (PyCFunction)cursor_iterprev_nodup, METH_VARARGS|METH_KEYWORDS},
     {"key", (PyCFunction)cursor_key, METH_NOARGS},
     {"last", (PyCFunction)cursor_last, METH_NOARGS},
+    {"last_dup", (PyCFunction)cursor_last_dup, METH_NOARGS},
     {"next", (PyCFunction)cursor_next, METH_NOARGS},
+    {"next_dup", (PyCFunction)cursor_next_dup, METH_NOARGS},
+    {"next_nodup", (PyCFunction)cursor_next_nodup, METH_NOARGS},
     {"prev", (PyCFunction)cursor_prev, METH_NOARGS},
+    {"prev_dup", (PyCFunction)cursor_prev_dup, METH_NOARGS},
+    {"prev_nodup", (PyCFunction)cursor_prev_nodup, METH_NOARGS},
     {"put", (PyCFunction)cursor_put, METH_VARARGS|METH_KEYWORDS},
     {"replace", (PyCFunction)cursor_replace, METH_VARARGS|METH_KEYWORDS},
     {"pop", (PyCFunction)cursor_pop, METH_VARARGS|METH_KEYWORDS},
     {"set_key", (PyCFunction)cursor_set_key, METH_O},
+    {"set_key_dup", (PyCFunction)cursor_set_key_dup, METH_VARARGS|METH_KEYWORDS},
     {"set_range", (PyCFunction)cursor_set_range, METH_O},
+    {"set_range_dup", (PyCFunction)cursor_set_range_dup, METH_VARARGS|METH_KEYWORDS},
     {"value", (PyCFunction)cursor_value, METH_NOARGS},
     {"_iter_from", (PyCFunction)cursor_iter_from, METH_VARARGS},
     {NULL, NULL}
@@ -2659,7 +2848,7 @@ trans_put(TransObject *self, PyObject *args, PyObject *kwds)
         int overwrite;
         int append;
         DbObject *db;
-    } arg = {{0, 0}, {0, 0}, 0, 1, 0, self->db};
+    } arg = {{0, 0}, {0, 0}, 1, 1, 0, self->db};
 
     static const struct argspec argspec[] = {
         {ARG_BUF, KEY_S, OFFSET(trans_put, key)},
@@ -2709,6 +2898,8 @@ trans_put(TransObject *self, PyObject *args, PyObject *kwds)
 
 static PyObject *
 make_cursor(DbObject *db, TransObject *trans);
+static PyObject *
+do_cursor_replace(CursorObject *self, MDB_val *key, MDB_val *val);
 
 /**
  * Transaction.replace() -> None|result
@@ -2735,38 +2926,13 @@ trans_replace(TransObject *self, PyObject *args, PyObject *kwds)
         return NULL;
     }
 
+    PyObject *ret = NULL;
     CursorObject *cursor = (CursorObject *) make_cursor(arg.db, self);
-    if(! cursor) {
-        return NULL;
+    if(cursor) {
+        ret = do_cursor_replace(cursor, &arg.key, &arg.value);
+        Py_DECREF(cursor);
     }
-
-    /* arg.val is updated if MDB_KEYEXIST. */
-    MDB_val newval = arg.value;
-    int flags = MDB_NOOVERWRITE;
-    self->mutations++;
-    int rc;
-    UNLOCKED(rc, mdb_cursor_put(cursor->curs, &arg.key, &arg.value, flags));
-    if(! rc) {
-        Py_DECREF((PyObject *) cursor);
-        Py_RETURN_NONE;
-    } else if(rc != MDB_KEYEXIST) {
-        Py_DECREF((PyObject *) cursor);
-        return err_set("mdb_put", rc);
-    }
-
-    PyObject *old = obj_from_val(&arg.value, 0);
-    if(! old) {
-        Py_DECREF((PyObject *) cursor);
-        return NULL;
-    }
-
-    UNLOCKED(rc, mdb_cursor_put(cursor->curs, &arg.key, &newval, 0));
-    Py_DECREF((PyObject *) cursor);
-    if(rc) {
-        Py_DECREF(old);
-        return err_set("mdb_put", rc);
-    }
-    return old;
+    return ret;
 }
 
 static int
